/*
    RAW input demo code
    - This code handles multiple HID devices individually.

    by MUTATE Systems 2018
*/

#define WIN32_LEAN_AND_MEAN
#include <windows.h>

#include <stdlib.h>
#include <malloc.h>
#include <memory.h>
#include <tchar.h>

#include "RawInput.h"
#include "RawInputListeners.h"

RawInputReceiver g_Receiver;
RiDetector<RiMouseListener> g_MouseDetector;
RiDetector<RiListener> g_KeyboardDetector;

const int NUM_RI_DEVICES = 2;

RiMouseState    g_MouseState[NUM_RI_DEVICES];
RiKeyboardState g_KeyboardState[NUM_RI_DEVICES];

inline const wchar_t* OnOff(bool b) {
    return b ? L"[+]" : L"[-]";
}

inline const wchar_t* strOK(bool b) {
    return b ? L"- OK" : L"";
}

inline const wchar_t* strTh(int n) {
    if(n == 1) return L"1st";
    if(n == 2) return L"2nd";
    if(n == 3) return L"3rd";

    static wchar_t sbuf[16];   
    swprintf_s(sbuf, 16, L"%dth", n);
    return sbuf;
}

LRESULT CALLBACK WndProc(HWND hWnd, UINT message, WPARAM wParam, LPARAM lParam)
{
    static bool bDetecting;
    const int MAX_SBUF = 200;
    wchar_t sbuf[MAX_SBUF];
    int x = 8, y = 8, h = 20;

	switch (message)
	{
    case WM_CREATE:
        SetTimer(hWnd, 1, 50, NULL);
        bDetecting = true;
        break;

    case WM_PAINT:
        {
            PAINTSTRUCT ps;
            HDC hdc = BeginPaint(hWnd, &ps);
            if(bDetecting) {

                for(int i = 0; i < NUM_RI_DEVICES; i ++) {
                    swprintf_s(sbuf, MAX_SBUF, L"Press any key of the %s player's keyboard. %s %08x", strTh(i + 1),
                      strOK(g_KeyboardDetector.getDeviceCount() > i), g_KeyboardDetector.getDeviceID(i));
                    TextOut(hdc, x, y, sbuf, wcslen(sbuf));
                    y += h;
                }

                for(int i = 0; i < NUM_RI_DEVICES; i ++) {
                    swprintf_s(sbuf, MAX_SBUF, L"Press any button of the %s player's mouse. %s %08x", strTh(i + 1),
                      strOK(g_MouseDetector.getDeviceCount() > i), g_MouseDetector.getDeviceID(i));
                    TextOut(hdc, x, y, sbuf, wcslen(sbuf));
                    y += h;
                }

                if(g_KeyboardDetector.getDeviceCount() >= NUM_RI_DEVICES && g_MouseDetector.getDeviceCount() >= NUM_RI_DEVICES) {
                    g_Receiver.initDeviceStatus();
                    for(int i = 0; i < NUM_RI_DEVICES; i ++) {
                        g_Receiver.addMouseListener(g_MouseDetector.getDeviceID(i), &g_MouseState[i]);
                        g_Receiver.addKeyboardListener(g_KeyboardDetector.getDeviceID(i), &g_KeyboardState[i]);
                    }
                    bDetecting = false;
                }
            }
            else {
                for(int i = 0; i < NUM_RI_DEVICES; i ++) {
                    swprintf_s(sbuf, MAX_SBUF, L"%s Player Mouse(%d,%d) L:%s R:%s M:%s B:%s F:%s Wheel:%d",
                        strTh(i + 1), g_MouseState[i].diffX(), g_MouseState[i].diffY(),
                        OnOff(g_MouseState[i].button(0)), OnOff(g_MouseState[i].button(1)),
                        OnOff(g_MouseState[i].button(2)), OnOff(g_MouseState[i].button(3)),
                        OnOff(g_MouseState[i].button(4)), 
                        g_MouseState[i].diffW() );
                    TextOut(hdc, x, y, sbuf, wcslen(sbuf));
                    y += h;
                }
                for(int i = 0; i < NUM_RI_DEVICES; i ++) {
                    swprintf_s(sbuf, MAX_SBUF, L"%s Player Keyboard W:%s A:%s S:%s D:%s", strTh(i + 1),
                        OnOff(g_KeyboardState[i].key('W')), OnOff(g_KeyboardState[i].key('A')),
                        OnOff(g_KeyboardState[i].key('S')), OnOff(g_KeyboardState[i].key('D')));
                    TextOut(hdc, x, y, sbuf, wcslen(sbuf));
                    y += h;
                }
            }
            EndPaint(hWnd, &ps);
        }
		break;

    case WM_TIMER:
        InvalidateRect(hWnd, NULL, TRUE);
        break;

    case WM_DESTROY:
		PostQuitMessage(0);
		break;
	default:
		return DefWindowProc(hWnd, message, wParam, lParam);
	}
	return 0;
}

int APIENTRY _tWinMain(HINSTANCE hInstance,
                     HINSTANCE hPrevInstance,
                     LPTSTR    lpCmdLine,
                     int       nCmdShow)
{
    WNDCLASSEX wcex = {0};
	wcex.cbSize = sizeof(WNDCLASSEX);
	wcex.style			= CS_HREDRAW | CS_VREDRAW;
	wcex.lpfnWndProc	= WndProc;
	wcex.hInstance		= hInstance;
    wcex.hCursor		= LoadCursor(NULL , IDC_ARROW);
    wcex.hbrBackground	= (HBRUSH)GetStockObject(WHITE_BRUSH);
	wcex.lpszClassName	= L"RawInputDemo";

    RegisterClassEx(&wcex);

    HWND hWnd = CreateWindow(wcex.lpszClassName, L"RawInput Demo", WS_OVERLAPPEDWINDOW,
        CW_USEDEFAULT, 0, CW_USEDEFAULT, 0, NULL, NULL, hInstance, NULL);

    if (!hWnd) return FALSE;

    RECT rc = {0};
    GetWindowRect(hWnd, &rc);
    MoveWindow(hWnd, rc.left, rc.top, 600, 400, FALSE);
    ShowWindow(hWnd, nCmdShow);
    UpdateWindow(hWnd);

    // Initialize RawInputReceiver
    g_Receiver.initialize();
    g_Receiver.setTargetWindow(hWnd);

    // Prepare the device detection
    g_Receiver.addKeyboardListener(RIDEV_DEFAULT, &g_KeyboardDetector);
    g_Receiver.addMouseListener(RIDEV_DEFAULT, &g_MouseDetector);

    MSG msg = {0};
	while (GetMessage(&msg, NULL, 0, 0))
	{
		TranslateMessage(&msg);
		DispatchMessage(&msg);
	}

	return (int) msg.wParam;
}
